@@ -2,14 +2,14 @@
 # Creation date: 2003-03-30 15:25:00
 # Authors: Don
 # Change log:
-# $Id: SelectLoop.pm,v 1.6 2004/07/01 06:37:11 don Exp $
+# $Id: SelectLoop.pm,v 1.10 2005/03/06 00:02:18 don Exp $
 
 use strict;
 
 {   package DBIx::Wrapper::SelectLoop;
 
     use vars qw($VERSION);
-    $VERSION = do { my @r=(q$Revision: 1.6 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
+    $VERSION = do { my @r=(q$Revision: 1.10 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
 
     use base 'DBIx::Wrapper::Statement';
 
@@ -18,25 +18,21 @@ use strict;
         my $self = { _query => $query, _exec_args => $exec_args,
                      _cur_row_count => 0
                    };
-        my $dbh = $parent->_getDatabaseHandle;
-        my $sth = $dbh->prepare($query);
-        unless ($sth) {
-            $parent->_printDbiError("\nQuery was '$query'\n");
-            return $parent->setErr(0, $DBI::errstr);
-        }
-        unless (scalar(@_) == 4) {
-            $exec_args = [];
-        }
-        $exec_args = [ $exec_args ] unless ref($exec_args);
-        unless ($sth->execute(@$exec_args)) {
-            $parent->_printDbiError("\nQuery was '$query'\n");
-            return $parent->setErr(1, $DBI::errstr);
-        }
-        
+
         bless $self, ref($proto) || $proto;
-        
-        $self->_setSth($sth);
         $self->_setParent($parent);
+
+        my ($sth, $rv, $r);
+        if (scalar(@_) == 4) {
+            ($sth, $rv, $r) = $parent->_getStatementHandleForQuery($query, $exec_args);
+        } else {
+            ($sth, $rv, $r) = $parent->_getStatementHandleForQuery($query);
+        }
+        
+        return $sth unless $sth;
+                        
+        $self->_setSth($sth);
+        $self->_setRequestObj($r);
         
         return $self;
     }
@@ -44,15 +40,36 @@ use strict;
     sub next {
         my ($self) = @_;
         my $sth = $self->_getSth;
-        $$self{_cur_row_count}++;
-        return $sth->fetchrow_hashref($self->_getParent()->getNameArg);
+        $self->{_cur_row_count}++;
+
+        my $r = $self->_getRequestObj;
+        $self->_getParent()->_runPreFetchHook($r);
+        $sth = $r->getStatementHandle;
+
+        my $result = $sth->fetchrow_hashref($self->_getParent()->getNameArg);
+
+        $r->setReturnVal($result);
+        $self->_getParent()->_runPostFetchHook($r);
+        $result = $r->getReturnVal;
+
+        return $result;
     }
 
     sub nextWithArrayRef {
         my ($self) = @_;
         my $sth = $self->_getSth;
-        $$self{_cur_row_count}++;
+        $self->{_cur_row_count}++;
+
+        my $r = $self->_getRequestObj;
+        $self->_getParent()->_runPreFetchHook($r);
+        $sth = $r->getStatementHandle;
+        
         my $row = $sth->fetchrow_arrayref;
+
+        $r->setReturnVal($row);
+        $self->_getParent()->_runPostFetchHook($r);
+        $row = $r->getReturnVal;
+        
         return [ @$row ] if $row;
         
         return undef;
@@ -110,6 +127,6 @@ DBIx::Wrapper::SelectLoop - Used by DBIx::Wrapper for SelectLoop methods
 
 =head1 VERSION
 
-$Id: SelectLoop.pm,v 1.6 2004/07/01 06:37:11 don Exp $
+$Id: SelectLoop.pm,v 1.10 2005/03/06 00:02:18 don Exp $
 
 =cut
